#include <cstring>
#include "pmf.h"

smat_t::format_t file_fmt = smat_t::PETSc;
bool do_shuffle = false;

void exit_with_help()
{
	printf(
	"Usage: omp-pmf-train [options] data_dir [model_filename]\n"
	"options:\n"
	"    -s type : set type of solver (default 0)\n"    
	"    	 0 -- CCDR1 with fundec stopping condition\n"    
	"    -k rank : set the rank (default 10)\n"    
	"    -n threads : set the number of threads (default 4)\n"    
	"    -l lambda : set the regularization parameter lambda (default 0.1)\n"    
	"    -t max_iter: set the number of iterations (default 5)\n"    
	"    -T max_iter: set the number of inner iterations used in CCDR1 (default 5)\n"    
	"    -e epsilon : set inner termination criterion epsilon of CCDR1 (default 1e-3)\n"     
	"    -p do_predict: do prediction or not (default 1)\n"    
	"    -q verbose: show information or not (default 0)\n"
	"    -N do_nmf: do nmf (default 0)\n"
	"    -S shuffle: random shuffle for rows and columns (default 0)\n"
	"    -f format: select input format (default 1)\n"
	"        0 -- plaintxt format\n"
	"        1 -- PETSc format\n"
	);
	exit(1);
}

pmf_parameter_t parse_command_line(int argc, char **argv, char *input_file_name, char *model_file_name)
{
	pmf_parameter_t param;   // default values have been set by the constructor 
	int i;

	// parse options
	for(i=1;i<argc;i++)
	{
		if(argv[i][0] != '-') break;
		if(++i>=argc)
			exit_with_help();
		switch(argv[i-1][1])
		{
			case 's':
				param.solver_type = atoi(argv[i]);
				break;

			case 'k':
				param.k = atoi(argv[i]);
				break;

			case 'n':
				param.threads = atoi(argv[i]);
				break;

			case 'l':
				param.lambda = atof(argv[i]);
				break;

			case 'r':
				param.rho = atof(argv[i]);
				break;

			case 't':
				param.maxiter = atoi(argv[i]);
				break;

			case 'T':
				param.maxinneriter = atoi(argv[i]);
				break;

			case 'e':
				param.eps = atof(argv[i]);
				param.eta0 = atof(argv[i]);
				break;

			case 'B':
				param.num_blocks = atoi(argv[i]);
				break;

			case 'm':
				param.lrate_method = atoi(argv[i]);
				break;

			case 'u':
				param.betaup = atof(argv[i]);
				break;

			case 'd':
				param.betadown = atof(argv[i]);
				break;

			case 'p':
				param.do_predict = atoi(argv[i]);
				break;

			case 'q':
				param.verbose = atoi(argv[i]);
				break;

			case 'N':
				param.do_nmf = atoi(argv[i]) == 1? true : false;
				break;

			case 'S':
				do_shuffle = atoi(argv[i]);
				break;

			case 'f':
				file_fmt = (smat_t::format_t) atoi(argv[i]);
				break;

			default:
				fprintf(stderr,"unknown option: -%c\n", argv[i-1][1]);
				exit_with_help();
				break;
		}
	}

	if (param.do_predict != 0) 
		param.verbose = 1;

	// determine filenames
	if(i>=argc)
		exit_with_help();

	strcpy(input_file_name, argv[i]);

	if(i<argc-1)
		strcpy(model_file_name,argv[i+1]);
	else
	{
		char *p = argv[i]+ strlen(argv[i])-1;
		while (*p == '/') 
			*p-- = 0;
		p = strrchr(argv[i],'/');
		if(p==NULL)
			p = argv[i];
		else
			++p;
		sprintf(model_file_name,"%s.model",p);
	}
	return param;
}


void run_ccdr1(pmf_parameter_t &param, const char* input_file_name, const char* model_file_name=NULL){
	FILE *model_fp = NULL;
	/*
	if(model_file_name) {
		model_fp = fopen(model_file_name, "wb");
		if(model_fp == NULL) {
			fprintf(stderr,"Error: can't open model file %s\n", model_file_name);
			exit(1);
		}
	}
	*/

	smat_t training_set, test_set;
	pmf_read_data(input_file_name, training_set, test_set, file_fmt);
	pmf_model_t model(training_set.rows, training_set.cols, param.k, pmf_model_t::COLMAJOR);

	// Random permutation for rows and cols of R for better load balancing
	size_t rows = training_set.rows, cols = training_set.cols;
	std::vector<unsigned> row_perm, inverse_row_perm;
	std::vector<unsigned> col_perm, inverse_col_perm;
	if(do_shuffle) {
		row_perm.resize(rows); inverse_row_perm.resize(rows);
		col_perm.resize(cols); inverse_col_perm.resize(cols);
		for(size_t r = 0; r < rows; r++) row_perm[r] = r;
		for(size_t c = 0; c < cols; c++) col_perm[c] = c;
		std::srand(0U);
		std::random_shuffle(row_perm.begin(), row_perm.end());
		std::random_shuffle(col_perm.begin(), col_perm.end());
		for(size_t r = 0; r < rows; r++) inverse_row_perm[row_perm[r]] = r;
		for(size_t c = 0; c < cols; c++) inverse_col_perm[col_perm[c]] = c;

		training_set.apply_permutation(row_perm, col_perm);
		test_set.apply_permutation(row_perm, col_perm);
	}

	puts("starts!");
	double time = omp_get_wtime();
	ccdr1(training_set, test_set, param, model);
	printf("Wall-time: %lg secs\n", omp_get_wtime() - time);

	if(model_fp) {
		if(do_shuffle) 
			model.apply_permutation(row_perm, col_perm);
		model.save(model_fp);
		fclose(model_fp);
	}
	return ;
}

int main(int argc, char* argv[]){
	char input_file_name[1024];
	char model_file_name[1024];
	pmf_parameter_t param = parse_command_line(argc, argv, input_file_name, model_file_name); 

	switch (param.solver_type){
		case CCDR1:
			run_ccdr1(param, input_file_name, model_file_name);
			break;
		default:
			fprintf(stderr, "Error: wrong solver type (%d)!\n", param.solver_type);
			break;
	}
	return 0;
}

