#include "pmf.h"

pmf_model_t::pmf_model_t(size_t rows_, size_t cols_, size_t k_, major_t major_type_, bool do_rand_init){
	rows = rows_;
	cols = cols_;
	k = k_;
	major_type = major_type_;
	if(do_rand_init)
		rand_init();
}

void pmf_model_t::mat_rand_init(mat_t &X, size_t m, size_t n, long seed) {
	val_type scale = 1./sqrt(k);
	srand48(seed);
	if(major_type == COLMAJOR) {
		X = mat_t(n, vec_t(m));
		for(size_t i = 0; i < m; i++)
			for(size_t j = 0; j < n; j++)
				X[j][i] = (val_type) (scale*(2*drand48()-1.0));
	} else { // major_type == ROWMAJOR
		X = mat_t(m, vec_t(n));
		for(size_t i = 0; i < m; i++)
			for(size_t j = 0; j < n; j++)
				X[i][j] = (val_type) (scale*(2*drand48()-1.0));
	}
}
void pmf_model_t::rand_init(long seed) {
	mat_rand_init(W, rows, k, seed);
	mat_rand_init(H, cols, k, seed+k);
}

void pmf_model_t::apply_permutation(const std::vector<unsigned> &row_perm, const std::vector<unsigned> &col_perm) {
	apply_permutation(row_perm.size()==rows? &row_perm[0]: NULL, col_perm.size()==cols? &col_perm[0] : NULL);
}
void pmf_model_t::apply_permutation(const unsigned *row_perm, const unsigned *col_perm) {

	if(major_type == COLMAJOR) {
		vec_t u(rows), v(cols);
		for(size_t t = 0; t < k; t++) {
			vec_t &Wt = W[t], &Ht = H[t];
			if(row_perm != NULL) {
				u = Wt;
				for(size_t r = 0; r < rows; r++) 
					Wt[r] = u[row_perm[r]];
			}
			if(col_perm != NULL) {
				v = Ht;
				for(size_t c = 0; c < cols; c++) 
					Ht[c] = v[col_perm[c]];
			}
		}
	} else { // major_type == ROWMAJOR
		if(row_perm != NULL) {
			mat_t buf = W;
			for(size_t r = 0; r < rows; r++)
				W[r] = buf[row_perm[r]];
		}
		if(col_perm != NULL) {
			mat_t buf = H;
			for(size_t c = 0; c < cols; c++)
				H[c] = buf[col_perm[c]];
		}
	}
}

void pmf_model_t::save(FILE *fp){
	save_mat_t(W, fp, major_type==ROWMAJOR);
	save_mat_t(H, fp, major_type==ROWMAJOR);
}

void pmf_model_t::load(FILE *fp, major_t major_type_){
	major_type = major_type_;
	W = load_mat_t(fp, major_type==ROWMAJOR);
	H = load_mat_t(fp, major_type==ROWMAJOR);
	rows = (major_type==ROWMAJOR)? W[0].size() : W.size();
	cols = (major_type==ROWMAJOR)? H[0].size() : H.size();
	k = (major_type==ROWMAJOR)? W[0].size(): W.size();
}

// Save a mat_t A to a file in row_major order.
// row_major = true: A is stored in row_major order,
// row_major = false: A is stored in col_major order.
void save_mat_t(mat_t A, FILE *fp, bool row_major){//{{{
	if (fp == NULL) 
		fprintf(stderr, "output stream is not valid.\n");
	long m = row_major? A.size(): A[0].size();
	long n = row_major? A[0].size(): A.size();
	fwrite(&m, sizeof(long), 1, fp);
	fwrite(&n, sizeof(long), 1, fp);
	double *buf = MALLOC(double, m*n);

	if (row_major) {
		size_t idx = 0;
		for(size_t i = 0; i < m; ++i) 
			for(size_t j = 0; j < n; ++j)
				buf[idx++] = A[i][j];
	} else {
		size_t idx = 0;
		for(size_t i = 0; i < m; ++i) 
			for(size_t j = 0; j < n; ++j)
				buf[idx++] = A[j][i];
	}
	fwrite(&buf[0], sizeof(double), m*n, fp);
	fflush(fp);
	free(buf);
}//}}}

// Load a matrix from a file and return a mat_t matrix 
// row_major = true: the returned A is stored in row_major order,
// row_major = false: the returened A  is stored in col_major order.
mat_t load_mat_t(FILE *fp, bool row_major){//{{{
	if (fp == NULL) 
		fprintf(stderr, "Error: null input stream.\n");
	long m, n; 
	if(fread(&m, sizeof(long), 1, fp) != 1)
		fprintf(stderr, "Error: wrong input stream.\n");
	if(fread(&n, sizeof(long), 1, fp) != 1)
		fprintf(stderr, "Error: wrong input stream.\n");
	double *buf = MALLOC(double, m*n);
	if(fread(&buf[0], sizeof(val_type), m*n, fp) != m*n)
		fprintf(stderr, "Error: wrong input stream.\n");
	mat_t A;
	if (row_major) {
		A = mat_t(m, vec_t(n));
		size_t idx = 0;
		for(size_t i = 0; i < m; ++i) 
			for(size_t j = 0; j < n; ++j)
				A[i][j] = buf[idx++];
	} else {
		A = mat_t(n, vec_t(m));
		size_t idx = 0;
		for(size_t i = 0; i < m; ++i) 
			for(size_t j = 0; j < n; ++j)
				A[j][i] = buf[idx++];
	}
	return A;
	free(buf);
}//}}}


// load utility for CCS RCS
void pmf_read_data(const char* srcdir, smat_t &training_set, smat_t &test_set, smat_t::format_t fmt){//{{{
	size_t m, n, nnz;
	char filename[1024], buf[1024], suffix[12];
	FILE *fp; 
	sprintf(filename,"%s/meta",srcdir);
	fp = fopen(filename,"r");
	if(fscanf(fp, "%lu %lu", &m, &n) != 2) {
		fprintf(stderr, "Error: corrupted meta in line 1 of %s\n", srcdir);
		return;
	}

	if(fscanf(fp, "%lu %s", &nnz, buf) != 2) {
		fprintf(stderr, "Error: corrupted meta in line 2 of %s\n", srcdir);
		return;
	}
	if(fmt == smat_t::TXT) 
		sprintf(suffix, "");
	else if(fmt == smat_t::PETSc)
		sprintf(suffix, ".petsc");
	else 
		printf("Error: fmt %d is not supported.", fmt);

	sprintf(filename,"%s/%s%s", srcdir, buf, suffix);
	training_set.load(m, n, nnz, filename, fmt);

	if(fscanf(fp, "%lu %s", &nnz, buf) != EOF){
		sprintf(filename,"%s/%s%s", srcdir, buf, suffix);
		test_set.load(m, n, nnz, filename, fmt);
	}
	fclose(fp);
	//val_type bias = training_set.get_global_mean(); training_set.remove_bias(bias); test_set.remove_bias(bias);
	return ;
}//}}}
