#ifndef _PMF_H_
#define _PMF_H_

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <algorithm>
#include <utility>
#include <map>
#include <queue>
#include <set>
#include <vector>
#include <cmath>
#include <assert.h>
#include <omp.h>

#ifdef _USE_FLOAT_
#define val_type float
#else 
#define val_type double
#endif

#include "smat.h"

typedef std::vector<val_type> vec_t;
typedef std::vector<vec_t> mat_t;

/* solver types*/
enum {CCDR1=0};



// pmf model 
class pmf_model_t {//{{{
	public:
		size_t rows, cols;
		size_t k;
		mat_t W, H;
		enum major_t {ROWMAJOR, COLMAJOR};
		int major_type;

		pmf_model_t(major_t major_type_=COLMAJOR): major_type(major_type){}
		pmf_model_t(size_t rows_, size_t cols_, size_t k_, major_t major_type_, bool do_rand_init=true);
		void rand_init(long seed=0L);
		void apply_permutation(const std::vector<unsigned> &row_perm, const std::vector<unsigned> &col_perm);
		void apply_permutation(const unsigned *row_perm=NULL, const unsigned *col_perm=NULL);
		void save(FILE *fp);
		void load(FILE *fp, major_t major_type_);
	private:
		void mat_rand_init(mat_t &X, size_t m, size_t n, long seed);
};//}}}

class pmf_parameter_t {//{{{
	public:
		int solver_type;
		long k;
		int threads;
		int maxiter, maxinneriter;
		double lambda;
		double rho;
		double eps;						// for the fundec stop-cond in ccdr1
		double eta0, betaup, betadown;  // learning rate parameters used in DSGD
		int lrate_method, num_blocks; 
		int do_predict, verbose;
		int do_nmf;  // non-negative matrix factorization
		pmf_parameter_t() {
			solver_type = CCDR1;
			k = 10;
			rho = 1e-3;
			maxiter = 5;
			maxinneriter = 5;
			lambda = 0.1;
			threads = 4;
			eps = 1e-3;
			/* options for dsgd
			eta0 = 1e-3; // initial eta0
			betaup = 1.05;
			betadown = 0.5;
			num_blocks = 30;  // number of blocks used in dsgd
			lrate_method = BOLDDRIVER;
			*/
			do_predict = 1;
			verbose = 0;
			do_nmf = 0;
		}
};//}}}


extern "C" {

	/* solvers*/
	void ccdr1(smat_t &training_set, smat_t &test_set, pmf_parameter_t &param, pmf_model_t &model);

	/*utility functions*/
	mat_t load_mat_t(FILE *fp, bool row_major=false);
	void save_mat_t(mat_t A, FILE *fp, bool row_major=false);
	void pmf_read_data(const char* srcdir, smat_t &training_set, smat_t &test_set, smat_t::format_t fmt = smat_t::TXT);
}


#endif // end of _PMF_H
