#ifndef SMAT_H
#define SMAT_H

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <algorithm>
#include <vector>
#include <cmath>
#include <assert.h>

#include <omp.h>
//#include "zlib/zlib.h"


#ifdef _USE_FLOAT_
#define val_type float
#else 
#define val_type double
#endif

#define MALLOC(type, size) (type*)malloc(sizeof(type)*(size))


class smat_t;
class entry_iterator_t; // iterator for files with (i,j,v) tuples
class smat_iterator_t; // iterator for nonzero entries in smat_t

// H = X*W, (X: m*n, W: n*k row-major, H m*k row major)
void smat_x_dmat(const smat_t &X, const val_type* W, const size_t k, val_type *H); 
// H = a*X*W + H0, (X: m*n, W: n*k row-major, H m*k row major)
void smat_x_dmat(const val_type a, const smat_t &X, const val_type* W, const size_t k, const val_type *H0, val_type *H);  

// Sparse matrix format CSC & CSR
class smat_t{
	private:
		bool mem_alloc_by_me;
		bool read_from_binary;
		unsigned char* binary_buf; 
		size_t binary_buf_len;
		const static int HeaderSize = 
			sizeof(size_t)+sizeof(size_t)+sizeof(size_t)+sizeof(size_t);
		void csr_to_csc();
		void csc_to_csr();
	public:
		size_t rows, cols;
		size_t nnz, max_row_nnz, max_col_nnz;
		val_type *val, *val_t;
		size_t *col_ptr, *row_ptr;
		unsigned *row_idx, *col_idx; 

		// filetypes for loading smat_t
		enum format_t {TXT=0, PETSc=1, BINARY=2, COMPRESSION=3};

		// Constructor and Destructor
		smat_t() : mem_alloc_by_me(false), read_from_binary(false), rows(0), cols(0), nnz(0){
		val=val_t=NULL; col_ptr=row_ptr=NULL, row_idx=col_idx=NULL;}
		smat_t(const smat_t& m){*this = m; mem_alloc_by_me = false; read_from_binary = false;}
		~smat_t(){ clear_space();}

		void clear_space();
		smat_t transpose();
		void apply_permutation(const std::vector<unsigned> &row_perm, const std::vector<unsigned> &col_perm);
		void apply_permutation(const unsigned *row_perm=NULL, const unsigned *col_perm=NULL);
		smat_t row_subset(const std::vector<unsigned> &subset);
		smat_t row_subset(const unsigned *subset, int subset_size);

		size_t nnz_of_row(unsigned i) const {return (row_ptr[i+1]-row_ptr[i]);}
		size_t nnz_of_col(unsigned i) const {return (col_ptr[i+1]-col_ptr[i]);}

		// smat-vector multiplication
		void Xv(const val_type *v, val_type *Xv); 
		void XTu(const val_type *u, val_type *XTu); 

		// IO methods
		void load_from_iterator(size_t _rows, size_t _cols, size_t _nnz, entry_iterator_t *entry_it);
		void load(size_t _rows, size_t _cols, size_t _nnz, const char *filename, format_t fmt);
		void load_from_PETSc(const char  *filename);
		void save_PETSc_to_file(const char *filename);
		void load_from_binary(const char *filename);
		void save_binary_to_file(const char *filename);

		// used for MPI verions
		void from_mpi(){
			mem_alloc_by_me = true;
			max_col_nnz = 0;
			for(size_t c = 0; c < cols; c++) 
				max_col_nnz = std::max(max_col_nnz, nnz_of_col(c));
		}
		val_type get_global_mean() const;
		void remove_bias(val_type bias=0);
};



/*-------------- Iterators -------------------*/

class rate_t{
	public:
		unsigned i, j; val_type v, weight;
		rate_t(int ii=0, int jj=0, val_type vv=0, val_type ww=1.0): i(ii), j(jj), v(vv), weight(ww){}
};

class entry_iterator_t {
	public:
		size_t nnz;
		virtual rate_t next() = 0;
};

// Iterator for files with (i,j,v) tuples
class file_iterator_t: public entry_iterator_t{
	public:
		file_iterator_t(size_t nnz_, const char* filename); 	
		~file_iterator_t(){ if (fp) fclose(fp); }
		rate_t next(); 
	private:
		size_t nnz;
		FILE *fp;
};


// smat_t iterator
class smat_iterator_t: public entry_iterator_t{
	public:
		enum {ROWMAJOR, COLMAJOR};
		// major: smat_iterator_t::ROWMAJOR or smat_iterator_t::COLMAJOR
		smat_iterator_t(const smat_t& M, int major = ROWMAJOR);
		~smat_iterator_t() {}
		rate_t next();
	private:
		unsigned *col_idx;
		size_t *row_ptr;
		val_type *val_t;
		size_t rows, cols, cur_idx;
		size_t cur_row;
};

// smat_t subset iterator
class smat_subset_iterator_t: public entry_iterator_t{
	public:
		enum {ROWMAJOR, COLMAJOR};
		// major: smat_iterator_t::ROWMAJOR or smat_iterator_t::COLMAJOR
		smat_subset_iterator_t(const smat_t& M, const unsigned *subset, size_t size, int major = ROWMAJOR);
		~smat_subset_iterator_t() {}
		rate_t next();
	private:
		unsigned *col_idx;
		size_t *row_ptr;
		val_type *val_t;
		size_t rows, cols, cur_idx;
		size_t cur_row;
		std::vector<unsigned>subset;
};

#endif
